import os

import oneiot_core.env as env
import oneiot_core.Parsers as Parsers

def set_variable(variable, value):
    # Set globally
    env = Parsers.EnvParser("/etc/environment")
    env.vars[variable] = value
    env.save()
    # Update in this session
    os.system(f'export {variable}="{value}"')

def unset_variable(variable):
    # Set globally
    env = Parsers.EnvParser("/etc/environment")
    if variable in env.vars:
        del env.vars[variable]
    env.save()
    # Update in this session
    os.system(f'unset {variable}')



import json
import asyncio
import threading

import websockets


class EventBus():

    def __init__(self, uri):
        self.uri = uri

        self.loop = asyncio.new_event_loop()
        # Create a queue of user inputs. There's no need to limit its size.
        self.inputs = asyncio.Queue(loop=self.loop)

        # Create a stop condition
        self.stop = self.loop.create_future()


    def connect(self):
        # Create an event loop that will run in a background thread.
        
        # Schedule the task that will manage the connection.
        asyncio.ensure_future(self.run_client(self.loop), loop=self.loop)

        # Start the event loop in a background thread.
        self.thread = threading.Thread(target=self.loop.run_forever)
        self.thread.start()
        
    def disconnect(self):
        self.loop.call_soon_threadsafe(self.stop.set_result, None)

        # Wait for the event loop to terminate.
        self.thread.join()

    async def run_client(self, loop):
        # Initiate the connection
        try:
            websocket = await websockets.connect(self.uri)
        except Exception as exc:
            # Probably should throw an error here
            return

        # Main loop
        try:
            while True:
                # Execute the incoming and outgoing tasks
                incoming = asyncio.ensure_future(websocket.recv())
                outgoing = asyncio.ensure_future(self.inputs.get())
                done, pending = await asyncio.wait([incoming, outgoing, self.stop], return_when=asyncio.FIRST_COMPLETED)

                # Cancel pending tasks to avoid leaking them.
                if incoming in pending:
                    incoming.cancel()
                if outgoing in pending:
                    outgoing.cancel()

                # Take action on the tasks
                # Incoming tasks
                if incoming in done:
                    message = incoming.result()
                    await self.consumer_handler(message)

                # Outgoing tasks
                if outgoing in done:
                    message = outgoing.result()
                    await websocket.send(message)
                    
                # Stop task
                if self.stop in done:
                    break

        finally:
            # Close the socket and running loop
            await websocket.close()
            asyncio.get_running_loop().stop()

    async def consumer_handler(self, message):
        #print(message)
        pass

    def send(self, id, data):
        message = json.dumps({
            'id': id,
            'data': data
        })
        self.loop.call_soon_threadsafe(self.inputs.put_nowait, message)


# a = EventBus("ws://192.168.2.111:1104")
# a.connect()
# a.send('hello.world', True)
# a.disconnect()



import os

defaults = {
    "ONEIOT_C_PORT": 1102,
    "ONEIOT_C_STATIC_IP": "192.168.4.1",
    "ONEIOT_C_NETWORK_INTERFACE": "wlan0",
    "ONEIOT_C_NETWORK_SSID": "OneIoT",
}

def var(name, default=None):
    if default == None:
        if name in defaults:
            default = defaults[name]
    return os.getenv(name) if os.getenv(name) is not None else default

def network_password():
    config = open("/etc/hostapd/hostapd.conf").read()
    for line in config.split("\n"):
        parts = line.split("=")
        if parts[0] == "wpa_passphrase":
            return parts[1]
    return None



import configparser
import os
import stat
import os.path as path

from pystemd.systemd1 import Unit

def get_status():
    # Get setup status
    service_file_created_status = get_service_file_created_status()
    service_is_enabled = check_service_is_enabled()
    # Get executing status
    service_is_running = check_service_is_running()
    return {"setup": [
            {"name": "Core Service unit file exists", "status": service_file_created_status},
            {"name": "Core Service starts on boot", "status": service_is_enabled},
        ],
        "execution": [
            {"name": "Core Service is running", "status": service_is_running},
        ]
    }

def get_service_file_created_status():
    if not path.exists("/lib/systemd/system/oneiotcore.service"):
        return False
    unit_file = configparser.RawConfigParser()
    unit_file.optionxform = str
    unit_file.read('/lib/systemd/system/oneiotcore.service')

    sections = unit_file.sections()
    result = 'Unit' in sections
    result = result and 'Service' in sections
    result = result and 'Install' in sections
    if not result:
        return False

    result = 'After' in unit_file['Unit']
    result = result and 'Type' in unit_file['Service']
    result = result and 'ExecStart' in unit_file['Service']
    result = result and 'WantedBy' in unit_file['Install']
    if not result:
        return False

    result = unit_file['Unit']['After'] == "multi-user.target"
    result = result and unit_file['Service']['Type'] == "idle"
    result = result and unit_file['Service']['ExecStart'] == "iot-core-serve"
    result = result and unit_file['Install']['WantedBy'] == "multi-user.target"

    st = os.stat("/lib/systemd/system/oneiotcore.service")
    result = result and st.st_mode & stat.S_IRUSR
    result = result and (st.st_mode & stat.S_IWUSR)
    result = result and (st.st_mode & stat.S_IRGRP)
    result = result and (st.st_mode & stat.S_IROTH)

    return result

def check_service_is_enabled():
    result = os.system('systemctl is-enabled oneiotcore.service > /dev/null')
    return result == 0

def check_service_is_running():
    result = os.system('systemctl status oneiotcore.service > /dev/null')
    return result == 0

def create_service_unit_file():
    if get_service_file_created_status():
        return
    print("Creating service unit file...")
    unit_file = configparser.RawConfigParser()
    unit_file.optionxform = str
    unit_file['Unit'] = {
        'Description': 'OneIoT Core',
        'After': 'multi-user.target'
    }
    unit_file['Service'] = {
        'Type': 'idle',
        'ExecStart': 'iot-core-serve'
    }
    unit_file['Install'] = {
        'WantedBy': 'multi-user.target'
    }
    with open('/lib/systemd/system/oneiotcore.service', 'w') as configfile:
        unit_file.write(configfile)
    print("... created!")

def start_core_on_boot():
    if check_service_is_enabled():
        return
    print("Enabling service on boot...")
    os.system('systemctl enable oneiotcore.service')
    print("... done!")

def start_core():
    if check_service_is_running():
        return
    print("Starting OneIoT Core...")
    os.system('systemctl start oneiotcore.service')
    print("... done!")



import os
import socketserver
from threading import Thread

import requests

import oneiot_core.env as env
import oneiot_core.webrepl_cli as webrepl_cli
import websocket, json

# connections = {}
# locks = {}

class service(socketserver.BaseRequestHandler):

    def handle(self):
        self.data = self.request.recv(1024).strip().decode("utf-8")
        command = self.data.split("\n")
        result = ""

        print(command)

        # if command[0] == "connect":
        #     result = self.connect(command[1:])
        if command[0] == "connect_test":
            result = self.connect_test(command[1:])
        elif command[0] == "send_to_device":
            result = self.send_to_device(command[1:])
        elif command[0] == "reset":
            result = self.reset(command[1:])
        elif command[0] == "reset_webrepl":
            result = self.reset_webrepl(command[1:])
        # elif command[0] == "disconnect":
        #     result = self.disconnect(command[1:])
        elif command[0] == "upload":
            result = self.upload(command[1:])
        elif command[0] == "heartbeat":
            result = self.heartbeat(command[1:])

        self.request.sendall(result)

    # args: [id, ip]
    # def connect(self, args):
    #     if args[0] not in connections:
    #         try:
    #             connections[args[0]] = websocket.WebSocket()
    #             connections[args[0]].connect("ws://" + args[1] + ":8266")
    #             connections[args[0]].recv()
    #             connections[args[0]].send("secret\n")
    #             connections[args[0]].recv()
    #             connections[args[0]].send("import user, json\r\n")
    #             for x in range(0,len("import user, json") + 2):
    #                 connections[args[0]].recv()
    #             locks[args[0]] = False
    #             return b'true'
    #         except Exception as e:
    #             if args[0] in connections:
    #                 del connections[args[0]]
    #             return(b'false\n' + str(e).encode())
    #     else:
    #         return b'false\nDevice already connected'

    # args: [id, ip]
    # def disconnect(self, args):
    #     if args[0] in connections:
    #         connections[args[0]].close()
    #         del locks[args[0]]
    #         del connections[args[0]]
    #         return b'true'
    #     else:
    #         return b'false\nDevice not connected'

    # args: [id, ip]
    def reset(self, args):
        result = requests.get(f'http://{args[1]}/sys/reset')
        return b'true'
        # if args[0] in connections:
        #     while locks[args[0]]:
        #         pass
        #     locks[args[0]] = True
        #     connections[args[0]].send("import machine\r\n")
        #     connections[args[0]].send("machine.reset()\r\n")
        #     self.disconnect([args[0]])
        #     self.connect([args[0], args[1]])
        #     locks[args[0]] = False
        #     return b'true'
        # else:
        #     return b'false\nDevice not connected'

    # args: [ip]
    def reset_webrepl(self, args):
        conn = websocket.WebSocket()
        conn.connect("ws://" + args[0] + ":8266")
        conn.send("secret\n")
        conn.send("import machine\r\n")
        conn.send("machine.reset()\r\n")
        conn.close()
        return b'true'

    # args: [id, ip, source, destination]
    def upload(self, args):
        id = args[0]
        ip = args[1]
        source = args[2]
        destination = args[3]
        try:
            requests.get(f'http://{args[1]}/sys/kill', timeout=1)
        except:
            pass
        try:
            webrepl_cli.main('secret', ip + ':' + destination, 'put', src_file=source)
            return b'true'
        except Exception as e:
            return b'false'

    # args: [id, ip]
    def connect_test(self, args):
        try:
            result = requests.get(f'http://{args[1]}/sys/test', timeout=1)
        except:
            return b'false'
        return result.content
        # if args[0] in connections:
        #     return b'true'
        # else:
        #     return b'false'

    # args: [id, ip, command, args (string)]
    def send_to_device(self, args):
        result = requests.post('http://' + args[1] + '/' + args[2], data=args[3], timeout=3)
        return result.content
        # args[2] = json.loads(args[2])
        # arg_string = ""
        # for arg in args[2]:
        #     arg_string += str(arg) + ","
        # command = "print(json.dumps(user." + args[1] + "(" + arg_string + ")))"
        #
        # while locks[args[0]]:
        #     pass
        # locks[args[0]] = True
        #
        # ws = connections[args[0]]
        #
        # ws.send(command + "\r\n")
        #
        # result = ""
        # addition = ""
        # while addition != ">>> ":
        #     addition = ws.recv()
        #     result += addition if addition != ">>> " else ""
        #
        # locks[args[0]] = False
        #
        # result_start = result.index("\r\n") + 2
        # result = result[result_start:]
        #
        # return result.encode()

    # args: []
    def heartbeat(self, args):
        return b'OK'

class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass

def main():
    HOST, PORT = "localhost", int(env.var("ONEIOT_C_PORT", 1102))
    server = ThreadedTCPServer((HOST, PORT), service)
    try:
        server.serve_forever()
    except Exception as e:
        server.shutdown()

if __name__ == "__main__":
    #main()



import os
import os.path as path

import netifaces
from pystemd.systemd1 import Unit
from clint.textui import prompt, validators

import oneiot_core.env as env
import oneiot_core.Parsers as Parsers

def get_status():
    # Get setup status
    interface_selected_status = get_interface_selected_status()
    static_ip_set_up_status = get_static_ip_set_up_status()
    hostapd_setup_status = get_hostapd_setup_status()
    dnsmasq_setup_status = get_dnsmasq_setup_status()
    # Get status
    static_ip_respected_status = get_static_ip_respected_status()
    hostapd_running_status = get_hostapd_running_status()
    dnsmasq_running_status = get_dnsmasq_running_status()
    return {"setup": [
            {"name": "Network Interface Selected", "status": interface_selected_status},
            {"name": "Static IP Set Up", "status": static_ip_set_up_status},
            {"name": "Hostapd Set Up", "status": hostapd_setup_status},
            {"name": "DNSMasq Set Up", "status": dnsmasq_setup_status},
        ],
        "status": [
            {"name": "Static IP Respected", "status": static_ip_respected_status},
            {"name": "HostAPD Running", "status": hostapd_running_status},
            {"name": "DNSMasq Running", "status": dnsmasq_running_status},
        ]
    }

def get_interface_selected_status():
    interfaces = netifaces.interfaces()
    return env.var("ONEIOT_C_NETWORK_INTERFACE") in interfaces

def get_static_ip_set_up_status():
    dhcpcd = Parsers.DHCPDParser("/etc/dhcpcd.conf")
    interface = env.var("ONEIOT_C_NETWORK_INTERFACE")
    static_ip = env.var("ONEIOT_C_STATIC_IP")
    if interface in dhcpcd.interfaces:
        static_present = False
        nohook_present = False
        for option in dhcpcd.interfaces[interface]:
            if option[0] == "static":
                static_present = option[1] == f"ip_address={static_ip}/24"
            if option[0] == "nohook":
                nohook_present = option[1] == f"wpa_supplicant"
        return static_present and nohook_present
    else:
        return False

def get_hostapd_setup_status():
    interface = env.var("ONEIOT_C_NETWORK_INTERFACE")
    ssid = env.var("ONEIOT_C_NETWORK_SSID")

    if not path.exists("/etc/hostapd/hostapd.conf"):
        return False
    if not path.exists("/etc/default/hostapd"):
        return False

    hostapd = Parsers.HostAPDParser("/etc/hostapd/hostapd.conf", "/etc/default/hostapd")

    options = hostapd.options
    print(options)
    print(ssid)
    result = options['interface'] == interface
    result = result and options['driver'] == 'nl80211'
    result = result and options['ssid'] == ssid
    result = result and options['hw_mode'] == 'g'
    result = result and options['channel'] == '6'
    result = result and options['ieee80211n'] == '1'
    result = result and options['wmm_enabled'] == '1'
    result = result and options['ht_capab'] == '[HT40][SHORT-GI-20][DSSS_CCK-40]'
    result = result and options['macaddr_acl'] == '0'
    result = result and options['auth_algs'] == '1'
    result = result and options['ignore_broadcast_ssid'] == '0'
    result = result and options['wpa'] == '2'
    result = result and options['wpa_key_mgmt'] == 'WPA-PSK'
    result = result and options['wpa_passphrase'] == env.network_password()
    result = result and len(env.network_password()) >= 8
    result = result and len(env.network_password()) <= 63
    result = result and options['rsn_pairwise'] == 'CCMP'

    options_master = hostapd.options_master
    if 'DAEMON_CONF' not in options_master:
        return False
    result = result and options_master['DAEMON_CONF'] == '"/etc/hostapd/hostapd.conf"'

    return result

def get_dnsmasq_setup_status():
    interface = env.var("ONEIOT_C_NETWORK_INTERFACE")
    static_ip = env.var("ONEIOT_C_STATIC_IP")
    static_ip_split = static_ip.split(".")

    if not path.exists("/etc/dnsmasq.conf"):
        return False

    dnsmasq = Parsers.DNSMasqParser("/etc/dnsmasq.conf")

    result = 'interface' in dnsmasq.option_dict
    result = result and 'server' in dnsmasq.option_dict
    result = result and 'dhcp-range' in dnsmasq.option_dict
    if not result:
        return result

    result = dnsmasq.option_dict['interface'] == interface
    result = result and dnsmasq.option_dict['server'] == '8.8.8.8'
    result = result and dnsmasq.option_dict['dhcp-range'] == f'{static_ip_split[0]}.{static_ip_split[1]}.{static_ip_split[2]}.2,{static_ip_split[0]}.{static_ip_split[1]}.{static_ip_split[2]}.254,255.255.255.0,24h'
    result = result and 'bind-interfaces' in dnsmasq.option_list
    result = result and 'domain-needed' in dnsmasq.option_list
    result = result and 'bogus-priv' in dnsmasq.option_list

    return result

def get_static_ip_respected_status():
    interface = env.var("ONEIOT_C_NETWORK_INTERFACE")
    ifconfig = netifaces.ifaddresses(interface)

    if netifaces.AF_INET in ifconfig:
        if len(ifconfig[netifaces.AF_INET]) > 0:
            return ifconfig[netifaces.AF_INET][0]['addr'] == env.var("ONEIOT_C_STATIC_IP")
        else:
            return False
    else:
        return False

def get_hostapd_running_status():
    unit = Unit(b'hostapd.service')
    unit.load()
    return unit.Unit.ActiveState == b'active'

def get_dnsmasq_running_status():
    unit = Unit(b'dnsmasq.service')
    unit.load()
    return unit.Unit.ActiveState == b'active'

def restart_services():
    hostapd = Unit(b'hostapd.service')
    hostapd.load()
    hostapd.Unit.Restart(b'replace')
    dnsmasq = Unit(b'dnsmasq.service')
    dnsmasq.load()
    dnsmasq.Unit.Restart(b'replace')

def setup_static_ip():
    if get_static_ip_set_up_status():
        return
    print("Setting up DHCPCD config...")
    dhcpcd = Parsers.DHCPDParser("/etc/dhcpcd.conf")
    interface = env.var("ONEIOT_C_NETWORK_INTERFACE")
    static_ip = env.var("ONEIOT_C_STATIC_IP")
    dhcpcd.modifyInterface(interface, [
        ['static', f'ip_address={static_ip}/24'],
        ['nohook', 'wpa_supplicant']
    ])
    dhcpcd.save()
    print(".. done!")

def setup_hostapd():
    if get_hostapd_setup_status():
        print(get_hostapd_setup_status())
        return

    print("Installing Hostapd...")
    os.system('apt-get -yq install hostapd')
    print("... done!")

    print("Setting up hostapd config...")

    interface = env.var("ONEIOT_C_NETWORK_INTERFACE")
    ssid = env.var("ONEIOT_C_NETWORK_SSID")
    psk = prompt.query('Network Password (8-63 alphanumeric chars)', validators=[validators.RegexValidator(regex=r"([a-z]|[0-9]){8,63}", message="Password must be 8 to 63 alphanumeric characters")])

    if not path.exists("/etc/hostapd"):
        os.mkdir("/etc/hostapd")

    hostapd = Parsers.HostAPDParser("/etc/hostapd/hostapd.conf", "/etc/default/hostapd")
    hostapd.set_options({
        'interface': interface,
        'driver': 'nl80211',
        'ssid': ssid,
        'hw_mode': 'g',
        'channel': '6',
        'ieee80211n': '1',
        'wmm_enabled': '1',
        'ht_capab': '[HT40][SHORT-GI-20][DSSS_CCK-40]',
        'macaddr_acl': '0',
        'auth_algs': '1',
        'ignore_broadcast_ssid': '0',
        'wpa': '2',
        'wpa_key_mgmt': 'WPA-PSK',
        'wpa_passphrase': psk,
        'rsn_pairwise': 'CCMP'
    })
    hostapd.save()

    hostapd.set_master_options({
        'DAEMON_CONF': '"/etc/hostapd/hostapd.conf"',
    })
    hostapd.save_master()

    os.system('update-rc.d hostapd enable')

    print(".. done!")

def setup_dnsmasq():
    if get_dnsmasq_setup_status():
        return

    print("Installing Dnsmasq...")
    os.system('apt-get -yq install dnsmasq')
    print("... done!")

    print("Setting up dnsmasq config...")

    interface = env.var("ONEIOT_C_NETWORK_INTERFACE")
    static_ip = env.var("ONEIOT_C_STATIC_IP")
    static_ip_split = static_ip.split(".")

    dnsmasq = Parsers.DNSMasqParser("/etc/dnsmasq.conf")
    dnsmasq.set_options({
        'interface': interface,
        'server': '8.8.8.8',
        'dhcp-range': f'{static_ip_split[0]}.{static_ip_split[1]}.{static_ip_split[2]}.2,{static_ip_split[0]}.{static_ip_split[1]}.{static_ip_split[2]}.254,255.255.255.0,24h',
    },
    [
        'bind-interfaces',
        'domain-needed',
        'bogus-priv'
    ])
    dnsmasq.save()

    print(".. done!")



import os.path as path

class DHCPDParser():

    def __init__(self, configPath):
        self.path = configPath
        if path.exists(configPath):
            self.raw = open(configPath).read()
        else:
            self.raw = ""
        self.modified = self.raw
        self.parse()


    def parse(self):
        lines = self.raw.split("\n")
        last_interface = None
        interfaces = {}
        for line in lines:
            if len(line) == 0:
                continue
            if line[0] == "#":
                continue
            splitLine = line.split(" ")
            if splitLine[0] == "interface":
                last_interface = splitLine[1]
                if splitLine[1] not in interfaces:
                    interfaces[splitLine[1]] = []
            elif splitLine[0][0:1] != "\t":
                last_interface = None
            elif last_interface is not None:
                if splitLine[0][0:1] == "\t":
                    splitLine[0] = splitLine[0][1:]
                    interfaces[last_interface].append(splitLine)
        self.interfaces = interfaces

    def modifyInterface(self, interface, options):
        result = ""
        in_interface = False
        for line in self.modified.split("\n"):
            if len(line) == 0:
                continue
            if line[0] == "#":
                continue
            splitLine = line.split(" ")
            if splitLine[0] == "interface":
                if splitLine[1] == interface:
                    in_interface = True
            elif splitLine[0][0:1] != "\t":
                in_interface = False
            if not in_interface:
                result += line + "\n"
        result += f"interface {interface}\n"
        for option in options:
            result += f"\t{option[0]}"
            for suboption in option[1:]:
                result += f" {suboption}"
            result += "\n"
        self.modified = result

    def save(self):
        f = open(self.path, "w")
        f.write(self.modified)
        f.close()

class HostAPDParser():

    def __init__(self, configPath, masterConfigPath):
        self.configPath = configPath
        if path.exists(configPath):
            self.raw = open(configPath).read()
        else:
            self.raw = ""
        self.masterConfigPath = masterConfigPath
        if path.exists(masterConfigPath):
            self.master_raw = open(masterConfigPath).read()
        else:
            self.master_raw = ""
        self.modified = self.raw
        self.master_modified = self.master_raw
        self.parse()
        self.parse_master()

    def parse(self):
        lines = self.raw.split("\n")
        self.options = {x.split("=")[0]: x.split("=")[1] for x in lines if len(x) > 0 and x[0] != "#"}

    def parse_master(self):
        lines = self.master_raw.split("\n")
        self.options_master = {x.split("=")[0]: x.split("=")[1] for x in lines if len(x) > 0 and x[0] != "#"}

    def set_options(self, options):
        result = ""
        for option in options:
            result += f"{option}={options[option]}\n"
        self.modified = result

    def set_master_options(self, options):
        result = ""
        for option in options:
            result += f"{option}={options[option]}\n"
        self.master_modified = result

    def save(self):
        f = open(self.configPath, "w")
        f.write(self.modified)
        f.close()

    def save_master(self):
        f = open(self.masterConfigPath, "w")
        f.write(self.master_modified)
        f.close()

class DNSMasqParser():

    def __init__(self, configPath):
        self.path = configPath
        if path.exists(configPath):
            self.raw = open(configPath).read()
        else:
            self.raw = ""
        self.modified = self.raw
        self.parse()

    def set_options(self, dict, list):
        result = ""
        for option in dict:
            result += f"{option}={dict[option]}\n"
        for option in list:
            result += f"{option}\n"
        self.modified = result

    def save(self):
        f = open(self.path, "w")
        f.write(self.modified)
        f.close()

    def parse(self):
        lines = self.raw.split("\n")
        self.option_dict = {}
        self.option_list = []
        for line in lines:
            split = line.split("=")
            if len(split) > 1:
                self.option_dict[split[0]]=split[1]
            else:
                self.option_list.append(split[0])

class EnvParser():

    def __init__(self, configPath):
        self.path = configPath
        if path.exists(configPath):
            self.raw = open(configPath).read()
        else:
            self.raw = ""
        self.parse()

    def parse(self):
        lines = self.raw.split("\n")
        self.vars = {x.split("=")[0]: x.split("=")[1] for x in lines if len(x) > 0 and x[0] != "#"}

    def save(self):
        f = open(self.path, "w")
        for var in self.vars:
            f.write(f"{var}={self.vars[var]}\n")
        f.close()
